/*************************************************************************
**************************************************************************
**                                                                      **
**  Project:   PS/2 Keyboard => NKC                                     **
**  Title:     PS/2 Keyboard => NKC Implementation                      **
**  File:      ps2keyboard.c                                            **
**  Date:      29.11.2008                                               **
**  Version:   0.96                                                     **
**  Plattform: WinAVR-20060421 & AVR Studio 4.12.498 & STK500           **
**                                                                      **
**  Created by Gerald Ebert                                             **
**                                                                      **
**************************************************************************
 Version   Description/Changes

   0.80    Project start with ATtiny2313.
   0.85  * At last the PS/2 Interface is functioning in both directions.
   0.86  + Watchdog timer implemented to avoid hangup on protocol errors.
   0.90  + Scancode -> ASCII/NKC translater added.
   0.91    Opimizing code, flash memory is running out (99,8 %).
   0.92  + Interface to NKC-KEY added, compatible to future NKC-KEY4.
   0.93    Opimizing code again, flash memory is running out (100,4 % !).
   0.94  * Made some bugfixes.
   0.95  + Activity indictor added, compatible to future NKC-KEY4.
   0.96  * Controller-Upgrade to ATmega48/88 and neccessary software changes made.
           Now there is enough flash memory for the future :)
         + Activity LED Function implemented.
         + PS2-Interface is blocked during initialization of the AVR.
         * Power-On-Sequence for keyboard changed.
         + Keycode buffer is also cleared when CTRL+ALT+DEL is detected.
         + ALT-Key combinations send an ESC-Sequence.
         + Trouble with $E1 extension keys on some keyboards corrected.
         + Handling Altgr-Keys and Windows-Keys.
         * Made some minor bugfixes. Debugging code for RS-232-terminal output.
   0.97  + Implementation of serial interface for mc-cpm computer
         + New configuration flags for interface selection

**************************************************************************
*************************************************************************/

#include "avr_defs.h"
#include "usart.h"
#include "fifo.h"
#include "ps2interface.h"

//========================================================================
// Device Configuration
//========================================================================

//#define TESTMODE
//#define INTERFACE_NKC      // Parallel interface for NKC
#define INTERFACE_MCCPM    // Serial interface for mc-cpm
//#define INVERT_OUTPUT      // Invert output byte before sending
#define MCCPM_BAUDRATE     USART_BAUD9K6
#define MCCPM_PARITYMODE   USART_PARITY_NONE
#define MCCPM_STOPBITS     USART_1STOP_BIT

#define KEYBOARDBUFFER_SIZE   32

//========================================================================
// Hardwiring the ATmega48/88
//========================================================================
/*
                   Parallel Interface for NKC
                             ____  ____
                   (RESET\) -|   \/   |- (ADC5/SCL) nc
              nc  (RxD/PD0) -|        |- (ADC4/SDA) nc
              nc  (TxD/PD1) -|        |- (ADC3)     nc
        PS/2-Clk (INT0/PD2) -|        |- (ADC2)     nc
      KEY5-Ready (INT1/PD3) -|        |- (ADC1)     nc
       PS/2-Data      (PD4) -|        |- (ADC0)     nc
                      (Vcc) -|        |- (GND)
                      (GND) -|        |- (Aref)
         KEY5-D6      (PB6) -|        |- (AVcc)
     KEY5-Strobe      (PB7) -|        |- (PB5)  KEY5-D5
        NKC-INT\      (PD5) -|        |- (PB4)  KEY5-D4
    Activity LED      (PD6) -|        |- (PB3)  KEY5-D3
      NKC-RESET\      (PD7) -|        |- (PB2)  KEY5-D2
         KEY4-D0      (PB0) -|________|- (PB1)  KEY5-D1


                  Serial Interface for mc-cpm
                             ____  ____
                   (RESET\) -|   \/   |- (ADC5/SCL) nc
              nc  (RxD/PD0) -|        |- (ADC4/SDA) nc
        KEY-Data  (TxD/PD1) -|        |- (ADC3)     nc
        PS/2-Clk (INT0/PD2) -|        |- (ADC2)     nc
      KEY5-Ready (INT1/PD3) -|        |- (ADC1)     nc
       PS/2-Data      (PD4) -|        |- (ADC0)     nc
                      (Vcc) -|        |- (GND)
                      (GND) -|        |- (Aref)
              nc      (PB6) -|        |- (AVcc)
              nc      (PB7) -|        |- (PB5)      nc
        NKC-INT\      (PD5) -|        |- (PB4)      nc
    Activity LED      (PD6) -|        |- (PB3)      nc
      NKC-RESET\      (PD7) -|        |- (PB2)      nc
              nc      (PB0) -|________|- (PB1)      nc


   * KEY5-Strobe is active high
*/

//========================================================================
// Defines
//========================================================================

#ifdef INTERFACE_MCCPM
  #undef TESTMODE            // No RS232 debugging when using serial interface for mc-cpm
#else
  #define INTERFACE_NKC      // Parallel interface is default
#endif

//------------------------------
// NKC-KEY interface

#define NKCREADY_PIN         (PIND & BV(PD3))
#define NKCSTROBE_OUT        BV(PB7)
#define NKCDATA_PORT         PORTB
// --- Handle strobe signal to latch keyboard data
#define NKC_STROBE           { NKCDATA_PORT |= NKCSTROBE_OUT;   NOP;   NKCDATA_PORT &= ~NKCSTROBE_OUT; }

//------------------------------
// External signals


#define NKCREADY_IN         BV(PD3)
#define NKCINT_OUT          BV(PD5)
#define NKCRESET_OUT        BV(PD7)
#define NKCSIG_PORT         PORTD
// --- Signal NKC new mouse data received (-> NKC-INT\)
#define NKC_INT_REQ         { NKCSIG_PORT &= ~NKCINT_OUT;  Wait_us(32);  NKCSIG_PORT |= NKCINT_OUT;   }

#define ActLED_Pin          BV(PD6)
#define ActLED_PORT         PORTD

// --- Set timer1 clock prescaler to 1/64. Set timer register to generate a 12,5 ms interrupt
#define RESETpin_INIT       { TCCR1B = 0x03; }
#define RESETpin_ENABLE     { CLI;  NKCSIG_PORT &= ~NKCRESET_OUT;  TCNT1 = 0xFFFF-0x061A;  TIFR1 |= BV(TOV1); \
                                    TIMSK1 |= BV(TOIE1);  SEI; }
#define RESETpin_DISABLE    { CLI;  NKCSIG_PORT |= NKCRESET_OUT;  TIMSK1 &= ~BV(TOIE1);  SEI;}

// --- Set timer2 clock prescaler to 1/1024. Set timer register to generate a 33 ms interrupt
#define ActLED_INIT         { TCCR2B = 0x07; }
#define ActLED_OFF          { ActLED_PORT &= ~ActLED_Pin; }
#define ActLED_ON           { ActLED_PORT |= ActLED_Pin; }
#define ActLED_ENABLE       { CLI;  ActLED_OFF   TCNT2 = 0;  TIFR2 |= BV(TOV2);  TIMSK2 |= BV(TOIE2);  SEI; }
#define ActLED_DISABLE      { CLI;  ActLED_ON;  TIMSK2 &= ~BV(TOIE2);  SEI;}

//------------------------------
// Host to keyboard commands

#define EXT_CODE            0xE0
#define EXT2_CODE           0xE1
#define SETLED_CODE         0xED
#define ECHO_CODE           0xEE
#define BREAK_CODE          0xF0
#define GET_KEYBOARD_ID     0xF2
#define SET_TYPEMATIC_RATE  0xF3
#define KBD_ENABLE          0xF4
#define KBD_DISABLE         0xF5

#define TYPEMATICRATE_30cps 0x00
#define TYPEMATICRATE_24cps 0x02
#define TYPEMATICRATE_21cps 0x04
#define TYPEMATICRATE_16cps 0x07
#define TYPEMATICRATE_12cps 0x0A
#define TYPEMATICRATE_10cps 0x0C
#define TYPEMATICRATE_8cps  0x0F
#define TYPEMATICRATE_6cps  0x12

#define TYPEMAT_DELAY_1s    0x60
#define TYPEMAT_DELAY_750ms 0x40
#define TYPEMAT_DELAY_500ms 0x20
#define TYPEMAT_DELAY_250ms 0x00

//------------------------------
// Keyboard LEDs

#define SCROLL_LOCK_LED     0x01
#define NUM_LOCK_LED        0x02
#define CAPS_LOCK_LED       0x04

//------------------------------
// Scancodes for special keys

#define LSHIFT_KEY          0x12
#define RSHIFT_KEY          0x59
#define SHIFTLOCK_KEY       0x58
#define LCTRL_KEY           0x14
#define LALT_KEY            0x11
#define NUM_KEY             0x77
#define SCROLL_KEY          0x7E

//------------------------------
// Codes for Keystates

#define LSHIFT_STATE        0x01
#define RSHIFT_STATE        0x02
#define SHIFTLOCK_STATE     0x04
#define SHIFT_STATE_MASK    0x07
#define LCTRL_STATE         0x08
#define RCTRL_STATE         0x10
#define CTRL_STATE_MASK     0x18
#define LALT_STATE          0x20
#define RALT_STATE          0x40
#define ALT_STATE_MASK      0x60
#define NUMLOCK_STATE       0x80
#define SCROLL_STATE        0x01

//========================================================================
// Local Variables
// (use volatile when variable is accessed from interrupts)
//========================================================================

static BOOL bKeyUpSeen;                          // Key Up (Break) code is previous data
static BOOL bExtenSeen, bExten2Seen;             // Extention code is previous data
static BYTE KeyStates, KeyStates2;               // States of Shift, Ctrl, Alt keys is down

static BYTE KeyboardBuffer [KEYBOARDBUFFER_SIZE]; // Keyboard buffer (32 bytes)
static tFIFO KeyboardFIFO;                        // Management struct for Keyboard buffer

static BOOL volatile KBDinitOK = FALSE;

//========================================================================
//========================================================================
// Debugger tools
//========================================================================
#if defined(TESTMODE)

//------------------------------------------------------------------------
// Converts a nibble to a digit
//------------------------------------------------------------------------

char Convert2Nibble(BYTE nibble)
{
   if ( nibble < 10 )
      return nibble + '0';
   else if ( nibble < 16 )
      return nibble-10 + 'A';
   return ' ';
}

//------------------------------------------------------------------------
// Send a byte as hex to USART
//------------------------------------------------------------------------

void SendScanCode(BYTE nScanCode)
{
    BYTE  hex;

   USART_Transmit('$');
   hex = Convert2Nibble((nScanCode>>4)&0x0F);
   USART_Transmit(hex);
   hex = Convert2Nibble(nScanCode&0x0F);
   USART_Transmit(hex);
}
#endif

//========================================================================
//========================================================================
// Interrupt Sevice Routines
//========================================================================
//========================================================================

//------------------------------------------------------------------------
// Default Interrupt Sevice Routine
//------------------------------------------------------------------------

EMPTY_INTERRUPT(__vector_default);

//------------------------------------------------------------------------
// TIMER1 Overflow Interrupt Sevice Routine
//------------------------------------------------------------------------

ISR(TIMER1_OVF_vect)
{
   RESETpin_DISABLE;                                              // Negate RESET Pin
   FIFO_init(&KeyboardFIFO, KeyboardBuffer, KEYBOARDBUFFER_SIZE); // Clear keyboard buffer
}

//------------------------------------------------------------------------
// TIMER2 Overflow Interrupt Sevice Routine
//------------------------------------------------------------------------

ISR(TIMER2_OVF_vect)
{
   ActLED_DISABLE;                               // Negate activity led
}

//========================================================================
//========================================================================
// Read data from internal EEPROM
//========================================================================

BYTE EEPROM_Read(BYTE address)
{
   
   while( (EECR & BV(EEPE)) );                   // Wait for completion of previous write
	EEAR = address;                               // Set up address register
	EECR |= BV(EERE);                             // Start eeprom read by writing EERE
   return(EEDR);                                 // Return data from data register
}

//========================================================================
//========================================================================
// Transfer a byte to FIFO
//========================================================================

void TransferData2FIFO(BYTE code)
{
    FIFO_in(&KeyboardFIFO, code);

    #if defined(TESTMODE)
    if ( code < ' ' )
    {
       code += 0x40;
       USART_Transmit('^');
    }
    USART_Transmit(code);
    #endif
}

//========================================================================
//========================================================================
// Transfer a byte to NKC
//========================================================================

void TransferData2NKC(void)
{
   BYTE   data;

   if ( NKCREADY_PIN && FIFO_isfilled(KeyboardFIFO) ) // Do we have data in keyboard buffer
   {                                                  // and NKC is ready to receive data?
#ifdef INVERT_OUTPUT
      data = ~FIFO_out(&KeyboardFIFO);           // Get next data byte and invert it
#else
      data = FIFO_out(&KeyboardFIFO);            // Get next data byte
#endif

#ifdef INTERFACE_NKC

      NKCDATA_PORT = data;                       // Transfer to NKC with strobe low
      NKC_STROBE;                                // Handle keyboard data strobe

#elif defined(INTERFACE_MCCPM)

      USART_Transmit(data);                      // Transfer to mc-cpm serial interface

#else
  #error "Interface configuration not supported."
#endif
   }
}

//========================================================================
//========================================================================
// Set Keyboard LEDs
//========================================================================

void SetKeyboardLEDs(void)
{
   BYTE LEDs = 0;

   if ( (KeyStates  & SHIFTLOCK_STATE) )         LEDs |= CAPS_LOCK_LED;
   if ( (KeyStates  & NUMLOCK_STATE) )           LEDs |= NUM_LOCK_LED;
   if ( (KeyStates2 & SCROLL_STATE) )            LEDs |= SCROLL_LOCK_LED;

   Send2PS2Device(SETLED_CODE);                  // Send set led command
   WaitForACKorData();                           // Wait until keyboard acknoledge received
   Send2PS2Device(LEDs);                         // Send led bits
   WaitForACKorData();                           // Wait until keyboard acknoledge received
}

//========================================================================
//========================================================================
// Handle Keyboard Power On Sequence
//========================================================================

void HandlePowerOn(void)
{
   Send2PS2Device(SETLED_CODE);                  // Send set led command
   WaitForACKorData();                           // Wait until keyboard acknoledge received
   Send2PS2Device(7);                            // Send led bits; all on
   WaitForACKorData();                           // Wait until keyboard acknoledge received

   Send2PS2Device(GET_KEYBOARD_ID);              // Send Get Keyboard ID command
   WaitForACKorData();                           // Wait until keyboard acknoledge received
   WaitForACKorData();                           // Wait until keyboard ID-Byte 1 (0xAB) received
   WaitForACKorData();                           // Wait until keyboard ID-Byte 2 (0x83) received

   Send2PS2Device(SETLED_CODE);                  // Send set led command
   WaitForACKorData();                           // Wait until keyboard acknoledge received
   Send2PS2Device(0);                            // Send led bits; all off
   WaitForACKorData();                           // Wait until keyboard acknoledge received
   
   Send2PS2Device(SET_TYPEMATIC_RATE);           // Send Set Typematic Rate/Delay command
   WaitForACKorData();                           // Wait until keyboard acknoledge received
   Send2PS2Device(TYPEMAT_DELAY_500ms|TYPEMATICRATE_16cps); // Send Typematic to 500ms / 16 cps (like Windows98)
   WaitForACKorData();                           // Wait until keyboard acknoledge received

   KeyStates = NUMLOCK_STATE;                    // Set numlock after keyboard startup
   SetKeyboardLEDs();

   KBDinitOK = TRUE;
}

//========================================================================
//========================================================================
// Scancode Decoder - Fills keybuffer for output to NKC
//========================================================================
//========================================================================

//------------------------------------------------------------------------
// Code table for scancodes without shift key - Standard keypad
//------------------------------------------------------------------------

static const PBYTE ScanCodesNormal [] =
{
//  -00-  -01-, -02-, -03-, -04-, -05-, -06-, -07-, -08-, -09-, -0A-, -0B-, -0C-, -0D-, -0E-, -0F-
    0x00, 0x92, 0x90, 0x8E, 0x8C, 0x8A, 0x8B, 0x95, 0x94, 0x93, 0x91, 0x8F, 0x8D, 0x09,  '^', 0x00,
//  -10-, -11-, -12-, -13-, -14-, -15-, -16-, -17-, -18-, -19-, -1A-, -1B-, -1C-, -1D-, -1E-, -1F-
    0x00, 0x00, 0x00, 0x00, 0x00,  'q',  '1', 0x00, 0x00, 0x00,  'y',  's',  'a',  'w',  '2', 0x00,
//  -20-, -21-, -22-, -23-, -24-, -25-, -26-, -27-, -28-, -29-, -2A-, -2B-, -2C-, -2D-, -2E-, -2F-
    0x00,  'c',  'x',  'd',  'e',  '4',  '3', 0x00, 0x00,  ' ',  'v',  'f',  't',  'r',  '5', 0x00,
//  -30-, -31-, -32-, -33-, -34-, -35-, -36-, -37-, -38-, -39-, -3A-, -3B-, -3C-, -3D-, -3E-, -3F-
    0x00,  'n',  'b',  'h',  'g',  'z',  '6', 0x00, 0x00, 0x00,  'm',  'j',  'u',  '7',  '8', 0x00,
//  -40-, -41-, -42-, -43-, -44-, -45-, -46-, -47-, -48-, -49-, -4A-, -4B-, -4C-, -4D-, -4E-, -4F-
    0x00,  ',',  'k',  'i',  'o',  '0',  '9', 0x00, 0x00,  '.',  '-',  'l', 0x7C,  'p', 0x7E, 0x00,
//  -50-, -51-, -52-, -53-, -54-, -55-, -56-, -57-, -58-, -59-, -5A-, -5B-, -5C-, -5D-, -5E-, -5F-
    0x00, 0x00, 0x7B, 0x00, 0x7D, 0x60, 0x00, 0x00, 0x00, 0x00, 0x0D,  '+', 0x00,  '#', 0x00, 0x00,
//  -60-, -61-, -62-, -63-, -64-, -65-, -66-, -67-,
    0x00,  '<', 0x00, 0x00, 0x00, 0x00, 0x7F, 0x00,
};

//------------------------------------------------------------------------
// Code table for scancodes with shift key - Standard keypad
//------------------------------------------------------------------------

static const PBYTE ScanCodesShift [] =
{
//  -00-  -01-, -02-, -03-, -04-, -05-, -06-, -07-, -08-, -09-, -0A-, -0B-, -0C-, -0D-, -0E-, -0F-
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x09, 0x40, 0x00,
//  -10-, -11-, -12-, -13-, -14-, -15-, -16-, -17-, -18-, -19-, -1A-, -1B-, -1C-, -1D-, -1E-, -1F-
    0x00, 0x00, 0x00, 0x00, 0x00,  'Q',  '!', 0x00, 0x00, 0x00,  'Y',  'S',  'A',  'W', 0x22, 0x00,
//  -20-, -21-, -22-, -23-, -24-, -25-, -26-, -27-, -28-, -29-, -2A-, -2B-, -2C-, -2D-, -2E-, -2F-
    0x00,  'C',  'X',  'D',  'E',  '$',  '#', 0x00, 0x00,  ' ',  'V',  'F',  'T',  'R',  '%', 0x00,
//  -30-, -31-, -32-, -33-, -34-, -35-, -36-, -37-, -38-, -39-, -3A-, -3B-, -3C-, -3D-, -3E-, -3F-
    0x00,  'N',  'B',  'H',  'G',  'Z',  '&', 0x00, 0x00, 0x00,  'M',  'J',  'U',  '/',  '(', 0x00,
//  -40-, -41-, -42-, -43-, -44-, -45-, -46-, -47-, -48-, -49-, -4A-, -4B-, -4C-, -4D-, -4E-, -4F-
    0x00,  ';',  'K',  'I',  'O',  '=',  ')', 0x00, 0x00,  ':',  '_',  'L', 0x5C,  'P',  '?', 0x00,
//  -50-, -51-, -52-, -53-, -54-, -55-, -56-, -57-, -58-, -59-, -5A-, -5B-, -5C-, -5D-, -5E-, -5F-
    0x00, 0x00, 0x5B, 0x00, 0x5D, 0x60, 0x00, 0x00, 0x00, 0x00, 0x0D,  '*', 0x00, 0x27, 0x00, 0x00,
//  -60-, -61-, -62-, -63-, -64-, -65-, -66-, -67-, -68-,
    0x00,  '>', 0x00, 0x00, 0x00, 0x00, 0x7F, 0x00, 0x00,
};

//------------------------------------------------------------------------
// Code table for scancodes without num key - Numeric keypad
//------------------------------------------------------------------------

static const PBYTE ScanCodesNumeric [] =
{
//                                                  -68-, -69-, -6A-, -6B-, -6C-, -6D-, -6E-, -6F-
                                                    0x00, 0x87, 0x00, 0x80, 0x86, 0x00, 0x00, 0x00,
//  -70-, -71-, -72-, -73-, -74-, -75-, -76-, -77-, -78-, -79-, -7A-, -7B-, -7C-, -7D-, -7E-, -7F-
    0x88, 0x89, 0x83, 0x00, 0x81, 0x82, 0x1B, 0x00, 0x00,  '+', 0x85,  '-',  '*', 0x84, 0x00, 0x00,
};

//------------------------------------------------------------------------
// Code table for scancodes with num key - Numeric keypad
//------------------------------------------------------------------------

static const PBYTE ScanCodesNumericNum [] =
{
//                                                  -68-, -69-, -6A-, -6B-, -6C-, -6D-, -6E-, -6F-
                                                    0x00,  '1', 0x00,  '4',  '7', 0x00, 0x00, 0x00,
//  -70-, -71-, -72-, -73-, -74-, -75-, -76-, -77-, -78-, -79-, -7A-, -7B-, -7C-, -7D-, -7E-, -7F-
     '0',  ',',  '2',  '5',  '6',  '8', 0x1B, 0x00, 0x00,  '+',  '3',  '-',  '*',  '9', 0x00, 0x00,
};

//------------------------------------------------------------------------
// Code table for scancodes of special keys (> 0x80 in other tables)
//------------------------------------------------------------------------

// NKC codes without ctrl key used for the editor
static const PBYTE SpecialCodes [] =
{
  0x13, 0x00,      // -80-    Move cursor left     / Arrow left
  0x04, 0x00,      // -81-    Move cursor right    / Arrow right
  0x05, 0x00,      // -82-    Move cursor up       / Arrow up
  0x18, 0x00,      // -83-    Move cursor down     / Arrow down
  0x12, 0x00,      // -84-    Page up              / Page up
  0x03, 0x00,      // -85-    Page down            / Page down
  0x11,  'S',      // -86-    Start of line        / Home
  0x11,  'D',      // -87-    End of Line          / End
  0x16, 0x00,      // -88-    Insert/Override      / Insert
  0x07, 0x00,      // -89-    Delete right char    / Delete

  0x0A, 0x00,      // -8A-    Help                 / F1
  0x0B,  'A',      // -8B-    Assembler            / F2
  0x0C, 0x00,      // -8C-    Repeat Search        / F3
  0x11,  'F',      // -8D-    Search               / F4
  0x0B,  'B',      // -8E-    Block begin          / F5
  0x0B,  'K',      // -8F-    Block end            / F6
  0x0B,  'C',      // -90-    Copy Block           / F7
  0x0B,  'V',      // -91-    Move Block           / F8
  0x10, 0x00,      // -92-    Characterset         / F9
  0x0B,  'X',      // -93-    End Editor           / F10
  0x0B,  'L',      // -94-    Load text block      / F11
  0x0B,  'S',      // -95-    Save text block      / F12

  0x00, 0x00,      // -96-                         / Left Window
  0x00, 0x00,      // -97-                         / Right Window
  0x00, 0x00,      // -98-                         / Menu
  0x00, 0x00,      // -99-                         / Power
  0x00, 0x00,      // -9A-                         / Sleep

  0x1B,  'S',      // -9B-                         / Scroll

};

// NKC codes with ctrl key used for the editor
static const PBYTE SpecialCodesCtrl [] =
{
  0x01, 0x00,      // -80-    Move word left       / Arrow left
  0x06, 0x00,      // -81-    Move word right      / Arrow right
  0x1A, 0x00,      // -82-    Move line up         / Arrow up
  0x17, 0x00,      // -83-    Move line down       / Arrow down
  0x11,  'E',      // -84-    Move to top of Page  / Page up
  0x11,  'X',      // -85-    Move to bottom of page / Page down
  0x11,  'R',      // -86-    Start of Text        / Home
  0x11,  'C',      // -87-    End of Text          / End
  0x0E, 0x00,      // -88-    Insert line          / Insert
  0x19, 0x00,      // -89-    Delete line          / Delete

  0x0A, 0x00,      // -8A-    Help                 / F1
  0x0B,  'A',      // -8B-    Assembler            / F2
  0x0C, 0x00,      // -8C-    Repeat Search        / F3
  0x11,  'A',      // -8F-    Replace              / F4
  0x0B,  'H',      // -8E-    Delete blockmarks    / F5
  0x0B,  'Y',      // -8D-    Deleteblock          / F6
  0x11,  'T',      // -90-    Split line           / F7
  0x11,  'V',      // -91-    Merge line           / F8
  0x1B,  'S',      // -92-    Scrollmode           / F9
  0x0B,  'Q',      // -93-    End Editor           / F10
  0x0B,  'R',      // -94-    Insert text from mem / F11
  0x00,  'W',      // -95-    Copy text to mem     / F12

  0x00, 0x00,      // -96-                         / Left Window
  0x00, 0x00,      // -97-                         / Right Window
  0x00, 0x00,      // -98-                         / Menu
  0x00, 0x00,      // -99-                         / Power
  0x00, 0x00,      // -9A-                         / Sleep

  0x1B,  'S',      // -9B-                         / Scroll

};

//------------------------------------------------------------------------
// Translates scancodes to NKC codes
//------------------------------------------------------------------------

void SetKeyState(BYTE nState)
{
   if ( bKeyUpSeen )   { KeyStates &= ~nState;   bKeyUpSeen = FALSE; }
   else                  KeyStates |= nState; 
   bExtenSeen = FALSE;
}

//------------------------------------------------------------------------

BYTE DecodeKey2ASCII(BYTE ScanCode)
{
   BYTE      NKCcode = 0;

   switch ( ScanCode )
   {
     case BREAK_CODE:
       bKeyUpSeen = TRUE;
       break;

     case EXT_CODE:
       bExtenSeen = TRUE;
       break;

     case LSHIFT_KEY:
       if ( bExtenSeen )    bKeyUpSeen = bExtenSeen = FALSE;     // Ignore fake shift key
       else                 SetKeyState(LSHIFT_STATE);
       break;

     case RSHIFT_KEY:
       SetKeyState(RSHIFT_STATE);
       break;

     case LCTRL_KEY:
       if ( bExtenSeen )    SetKeyState(RCTRL_STATE);
       else if ( bExten2Seen ) 
                          { bKeyUpSeen = bExten2Seen = FALSE; }  // Ignore break key
       else                 SetKeyState(LCTRL_STATE);
       break;

     case LALT_KEY:
       if ( bExtenSeen )    SetKeyState(RALT_STATE);
       else                 SetKeyState(LALT_STATE);
       break;

     case SHIFTLOCK_KEY:
       if ( bKeyUpSeen )    bKeyUpSeen = FALSE;
       else               { KeyStates ^= SHIFTLOCK_STATE;  SetKeyboardLEDs(); }
       break;

     case NUM_KEY:
       if ( bKeyUpSeen )    bKeyUpSeen = FALSE;
       else               { KeyStates ^= NUMLOCK_STATE;  SetKeyboardLEDs();   }
       break;

     case SCROLL_KEY:
       if ( bKeyUpSeen )    bKeyUpSeen = FALSE;
       else               { KeyStates2 ^= SCROLL_STATE;  SetKeyboardLEDs();  NKCcode = 0x9B; }
       break;

     case EXT2_CODE:
       bExten2Seen = TRUE;
       break;

     default:
       if ( ScanCode == 0x83 )   ScanCode = 0x02;     // Handle F7 key. Remap scancode
       if ( ScanCode == 0x78 )   ScanCode = 0x08;     // Handle F11 key. Remap scancode
       if ( bKeyUpSeen )  { bKeyUpSeen = bExtenSeen = bExten2Seen = FALSE; }
       // Handle keys with Extension Code
       else if ( bExtenSeen )
       {
          bExtenSeen = FALSE;
		    if ( ScanCode == 0x4A )  NKCcode = '/';     // '/' key on numeric keypad
          if ( ScanCode == 0x1F )  NKCcode = 0x96;    // Handle Left Window key
          if ( ScanCode == 0x27 )  NKCcode = 0x97;    // Handle Right Window key
          if ( ScanCode == 0x2F )  NKCcode = 0x98;    // Handle Menu key
          if ( ScanCode == 0x37 )  NKCcode = 0x99;    // Handle Power key
          if ( ScanCode == 0x3F )  NKCcode = 0x9A;    // Handle Sleep key
		    if ( ScanCode == 0x5A )                     // 'Enter' key on numeric keypad
		       NKCcode = GET_PBYTE(ScanCodesNormal+ScanCode);
          if ( ScanCode > 0x67 && ScanCode < 0x80 )
             NKCcode = GET_PBYTE(ScanCodesNumeric+ScanCode-0x68);
       }
       else if ( bExten2Seen )
       {
          bExten2Seen = FALSE;
       }
       // Handle Standard keypad
       else if ( ScanCode < 0x68 )
       {
          // Handle keys with Ctrl
          if ( (KeyStates & CTRL_STATE_MASK) )
          {
             NKCcode = GET_PBYTE(ScanCodesNormal+ScanCode);
             if ( NKCcode >= 0x60 && NKCcode <= 0x7F )
                NKCcode &= 0x1F;                      // Create CTRL-Codes
             else if (NKCcode < 0x80 )           
                NKCcode = 0;                          // Everything else but special codes must be zero
          }
          // Handle keys with Altgr
          else if ( (KeyStates & RALT_STATE) )
          {
             if ( ScanCode == 0x15 )   NKCcode = 0x40;     // '@'
             if ( ScanCode == 0x24 )   NKCcode = 0x00;     // Euro
             if ( ScanCode == 0x3A )   NKCcode = 0x00;     // ''
             if ( ScanCode == 0x3D )   NKCcode = 0x7B;     // '{'
             if ( ScanCode == 0x3E )   NKCcode = 0x5B;     // '['
             if ( ScanCode == 0x45 )   NKCcode = 0x7D;     // '}'
             if ( ScanCode == 0x46 )   NKCcode = 0x5D;     // ']'
             if ( ScanCode == 0x4E )   NKCcode = 0x5C;     // Backslash
             if ( ScanCode == 0x5B )   NKCcode = 0x7E;     // '~'
             if ( ScanCode == 0x61 )   NKCcode = 0x7C;     // '|'
          }
          // Handle keys with Shift
          else if ( (KeyStates & SHIFT_STATE_MASK) )
             NKCcode = GET_PBYTE(ScanCodesShift+ScanCode);
          else
             NKCcode = GET_PBYTE(ScanCodesNormal+ScanCode);
       }
       // Handle Numeric keypad
       else if ( ScanCode < 0x80 )
       {
          ScanCode -= 0x68;
          if ( (KeyStates & NUMLOCK_STATE) )
             NKCcode = GET_PBYTE(ScanCodesNumericNum+ScanCode);
          else
             NKCcode = GET_PBYTE(ScanCodesNumeric+ScanCode);
       }
       break;
   }
   #if defined(TESTMODE)
   SendScanCode(ScanCode);
   #endif
   return NKCcode;
}

//------------------------------------------------------------------------
// Send codes to NKC
//------------------------------------------------------------------------


void Send2NKC(BYTE NKCcode)
{
   BYTE      a, b;
   LPCPBYTE  cp;

   if ( NKCcode )
   {
      a = (KeyStates & CTRL_STATE_MASK);
      b = (KeyStates & LALT_STATE);
      if ( NKCcode < 0x80 )
      {  // Handle standard code table
         if ( a && (! b) && NKCcode >= 0x20 )
            //  Only ALT pressed
            TransferData2FIFO(0x1B);             // Send ESC first
         TransferData2FIFO(NKCcode);
      }
      else
      {  // Handle NKC extension code table
         if ( a && b )
         { // Ctrl+ALT pressed
           if ( NKCcode == 0x89 )
           {  // Ctrl+ALT+Del detected
              // Reset NKC - CPU (-> NKC-RESET\)
              RESETpin_ENABLE;
              return;
           }
           if ( NKCcode == 0x1B )
           {  // Ctrl+ALT+ESC detected
              // Clear Keyboard buffer
              FIFO_init(&KeyboardFIFO, KeyboardBuffer, KEYBOARDBUFFER_SIZE);
              return;
           }
         }
         else
         {
            NKCcode <<= 1;                       // => NKCcode -= 0x80; NKCcode *= 2;
            cp = (LPCPBYTE)SpecialCodes;
            if ( a )
               cp = (LPCPBYTE)SpecialCodesCtrl;
            if ( (a = GET_PBYTE(cp+NKCcode)) )   // Must not be zero
            {
               TransferData2FIFO(a);             // Add byte to keyboard buffer
               if ( (a = GET_PBYTE(cp+NKCcode+1)) ) // Must not be zero
                  TransferData2FIFO(a);          // Add byte to keyboard buffer
            }
         }
      }
      // Signal NKC new keyboard data received (-> NKC-INT\)
      NKC_INT_REQ;
   }
}

//========================================================================
//========================================================================
// MAIN
//========================================================================

int main (void)
{
   BYTE temp;
   BOOL bRepeatSend = FALSE;
   BOOL bKBinitOK = FALSE;

   STARTUP_INIT;                                 // AVR startup sequence
   PS2DATA_PORT &= ~(PS2DATA_CLK );              // Set clock to low to signal keyboard that host is buzy

   Set_Sleep_Mode(SLEEP_MODE_IDLE);              // Initialize Sleep Mode for AVR
   WDTI_IntEnable(WDT_250MS);                    // Enable Watchdog Timer Interrupt
   WDerrorSeen = FALSE;
   #if defined(TESTMODE)
   USART_Init(USART_BAUD38K, USART_PARITY_EVEN, USART_1STOP_BIT); // Initialize USART for testing
   #endif
   #if defined (INTERFACE_MCCPM)
   USART_Init(MCCPM_BAUDRATE, MCCPM_PARITYMODE, MCCPM_STOPBITS); // Initialize USART for mc-cpm serial interface
   #endif
   RESETpin_INIT;                                // Initialize TIMER1 for system reset pin
   ActLED_INIT;                                  // Initialize TIMER2 for activation led blinking
   ActLED_OFF;                                   // Switch LED off

   NKCDATA_PORT &= ~NKCSTROBE_OUT;               // Set strobe to low level so NKC can latch the reset of key data port
   PS2DATA_DDR  &= ~(NKCREADY_IN);               // Set ready pin as input
   PS2DATA_PORT |= NKCREADY_IN;                  // Set pullup at ready pin
   DETECT0_FALLINGEDGE;                          // Set INT0 on falling edge
   DETECT1_RISINGEDGE;                           // Set INT1 on rising edge
   ENABLE_INT0;                                  // Enable external INT0
   ENABLE_INT1;                                  // Enable external INT1

   FIFO_init(&KeyboardFIFO, KeyboardBuffer, KEYBOARDBUFFER_SIZE); // Initialize keyboard buffer

   PS2DATA_PORT |= (PS2DATA_CLK | PS2DATA_OUT);  // Set clock and data to high to signal keyboard bus is idle
   PS2DATA_DDR  &= ~(PS2DATA_CLK | PS2DATA_OUT); // Set clock and data pin as input
   PS2DATA_PORT &= ~(PS2DATA_CLK | PS2DATA_OUT); // Set no pullup for clock and data pin

   #if defined(TESTMODE)
   USART_Transmit('*');
   USART_Transmit('S');
   USART_Transmit('*');
   #endif

   SEI;                                          // Enable all used ISR

   if ( RestartCause == BV(EXTRF) )              // External reset detected at startup
   {
      Send2PS2Device(RESET_CODE);                // Send Keyboard Reset command
      #if defined(TESTMODE)
      USART_Transmit('!');
      #endif
      WaitForACKorData();                        // Wait until keyboard acknoledge received
      WDerrorSeen = FALSE;                       // Reset WDT overflow signal
   }

   while ( TRUE )  // LOOP FOREVER
   {
      // Wait until data byte received or watchdog
      while ( PS2State != PS2STATE_DATA_RECEIVED )
      {
         SLEEP;  WDT;
         TransferData2NKC();                     // Transfer next data byte to NKC
         if ( WDerrorSeen )
         {
            #if defined(TESTMODE)
            USART_Transmit(':');
            #endif
            WDerrorSeen = FALSE;                 // Reset WDT overflow signal
            continue;
         }
      }
      PS2State = PS2STATE_IDLE;                  // Reset State to idle.

      temp = rParityBits & 0x01;                 // Check parity bit
      if ( temp != rParity && (! bRepeatSend) )
      {                                          // Check failed !
         Send2PS2Device(RESEND_CODE);            // Send repeat last byte command on fail
         bRepeatSend = TRUE;                     // Only one reapeat command per data byte
      }
      else
      {
         bRepeatSend = FALSE;                    // Reset repeat indicator
         if ( rData == PWRON_CODE )              // Keyboard Power On received
         {
            ActLED_OFF;                          // Switch LED off
            HandlePowerOn();                     // Handle Keyboard Power On Sequence
            ActLED_ON;                           // Switch LED on
			bKBinitOK = TRUE;
            #if defined(TESTMODE)
            USART_Transmit('&');
            #endif
         }
         else
         {
            ActLED_ENABLE;                       // Signal keyboard activity
            temp = DecodeKey2ASCII(rData);       // Decode scancode to ASCII code and Specials
            Send2NKC(temp);                      // Send it to NKC
         }
      }
      TransferData2NKC();                        // Transfer next data byte to NKC
   }
}

//========================================================================
