/*************************************************************************
**************************************************************************
**                                                                      **
**  Project:   PS/2 Mouse => NKC                                        **
**  Title:     PS/2 Mouse => NKC Implementation                         **
**  File:      ps2mouse.c                                               **
**  Date:      23.02.2007                                               **
**  Version:   0.95                                                     **
**  Plattform: WinAVR-20060421 & AVR Studio 4.12.498 & STK500           **
**                                                                      **
**  Created by Gerald Ebert                                             **
**                                                                      **
**************************************************************************
 Version   Description/Changes

   0.80    Project start. PS/2 Interface and management taken from keyboard project
   0.90    Scancode -> Mouse position decoder with error management added
   0.91    Interface to NKC-HCOPY added, compatible to future NKC-KEY4
   0.94    Made some bugfixes
   0.95    Activity LED controller added, compatible to future NKC-KEY4
   0.96    Watchdog functionality corrected.
           Pin Layout corrected for KEY5.
           Scroll mouse function added. Supported by KEY5.


**************************************************************************
*************************************************************************/

#include "avr_defs.h"
#include "ps2interface.h"

//#define TESTMODE

//========================================================================
// Hardwiring the ATtiny2313
//========================================================================
/*
                             ____  ____
                   (RESET\) -|   \/   |- (Vcc)
              nc  (RxD/PD0) -|        |- (PB7) KEY5-mouseDown
              nc  (TxD/PD1) -|        |- (PB6) KEY5-mouseUp
   KEY5-mouseZdn      (PA1) -|        |- (PB5) KEY5-mouseRight
   KEY5-mouseZup      (PA0) -|        |- (PB4) KEY5-mouseLeft
        PS/2-Clk (INT0/PD2) -|        |- (PB3) KEY5-mouseButton4
              nc (INT1/PD3) -|        |- (PB2) KEY5-mouseButton3 (middle)
       PS/2-Data      (PD4) -|        |- (PB1) KEY5-mouseButton2 (right)
        NKC-INT\      (PD5) -|        |- (PB0) KEY5-mouseButton1 (left)
                      (GND) -|________|- (PD6) Activity LED

*/

//========================================================================
// Defines
//========================================================================

//------------------------------
// NKC-KEY interface

#define NKC_MOUSELEFT        BV(PB4)
#define NKC_MOUSERIGHT       BV(PB5)
#define NKC_MOUSEUP          BV(PB6)
#define NKC_MOUSEDOWN        BV(PB7)
#define NKC_MOUSEZUP         BV(PA0)
#define NKC_MOUSEZDOWN       BV(PA1)
#define NKC_MOUSEBUTTON1     BV(PB0)
#define NKC_MOUSEBUTTON2     BV(PB1)
#define NKC_MOUSEBUTTON3     BV(PB2)
#define NKC_MOUSEBUTTON4     BV(PB3)
#define NKC_MOUSEBUTTONMASK  (NKC_MOUSEBUTTON1 | NKC_MOUSEBUTTON2 | NKC_MOUSEBUTTON3 | NKC_MOUSEBUTTON4)
#define NKCDATA_PORT         PORTB
// --- Make a short pulse to increment the external counters
#define NKC_MOUSELEFT_COUNT  { NKCDATA_PORT |= NKC_MOUSELEFT;   NOP;   NKCDATA_PORT &= ~NKC_MOUSELEFT;  }
#define NKC_MOUSERIGHT_COUNT { NKCDATA_PORT |= NKC_MOUSERIGHT;  NOP;   NKCDATA_PORT &= ~NKC_MOUSERIGHT; }
#define NKC_MOUSEUP_COUNT    { NKCDATA_PORT |= NKC_MOUSEUP;     NOP;   NKCDATA_PORT &= ~NKC_MOUSEUP;    }
#define NKC_MOUSEDOWN_COUNT  { NKCDATA_PORT |= NKC_MOUSEDOWN;   NOP;   NKCDATA_PORT &= ~NKC_MOUSEDOWN;  }

//------------------------------
// External signals

#define NKCINT_OUT          BV(PD5)
#define NKCSIG_PORT         PORTD
// --- Signal NKC new mouse data received (-> NKC-INT\)
#define NKC_INT_REQ         { NKCSIG_PORT &= ~NKCINT_OUT;    Wait_us(32);   NKCSIG_PORT |= NKCINT_OUT; }

#define ActLED_Pin          BV(PD6)
#define ActLED_PORT         PORTD
// --- Set timer0 clock prescaler to 1/1024. Set timer register to generate a 33 ms interrupt
#define ActLED_INIT         { TCCR0B = 0x05; }
#define ActLED_OFF          { ActLED_PORT &= ~ActLED_Pin; }
#define ActLED_ON           { ActLED_PORT |= ActLED_Pin; }
#define ActLED_ENABLE       { CLI;  ActLED_OFF   TCNT0 = 0;  TIFR |= BV(TOV0);  TIMSK |= BV(TOIE0);  SEI; }
#define ActLED_DISABLE      { CLI;  ActLED_ON;  TIMSK &= ~BV(TOIE0);  SEI;}

//------------------------------
// Host to mouse commands

#define SET_RESOLUTION       0xE8
#define ECHO_CODE            0xEE
#define GET_DEVICE_ID        0xF2
#define SET_SAMPLE_RATE      0xF3
#define DATA_REPORT_ENABLE   0xF4
#define DATA_REPORT_DISABLE  0xF5
#define SET_DEFAULT_CODE     0xF6

#define RESOLUTION1cmm       0x00
#define RESOLUTION2cmm       0x01
#define RESOLUTION4cmm       0x02
#define RESOLUTION8cmm       0x03

#define SAMPLERATE_10sps       10
#define SAMPLERATE_20sps       20
#define SAMPLERATE_40sps       40
#define SAMPLERATE_60sps       60
#define SAMPLERATE_80sps       80
#define SAMPLERATE_100sps     100
#define SAMPLERATE_200sps     200

//========================================================================
// Local Variables
// (use volatile when variable is accessed from interrupts)
//========================================================================

static BYTE mousePacket [3];                     // Buffer for mouse packet data
static BYTE mpCounter;                           // Counter for incoming packet data

static char XLpos, XRpos, YUpos, YDpos;          // The current relative mouse position
static BYTE MouseButtons;                        // Current mouse button states

//========================================================================
//========================================================================
// Interrupt Sevice Routines
//========================================================================
//========================================================================

//------------------------------------------------------------------------
// Default Interrupt Sevice Routine
//------------------------------------------------------------------------

EMPTY_INTERRUPT(__vector_default);

//------------------------------------------------------------------------
// TIMER0 Overflow Interrupt Sevice Routine
//------------------------------------------------------------------------

ISR(TIMER0_OVF_vect)
{
   ActLED_DISABLE;                              // Switch activity led off
}

//========================================================================
//========================================================================
// USART
//========================================================================

#if defined(TESTMODE)

#include "USART.inc"                             // This is a terrable way to include code but i am lazy here.

#endif

//========================================================================
//========================================================================
// Handle Mouse Power On Sequence
//========================================================================

void HandlePowerOn(void)
{
   WaitForACKorData();                           // Wait until mouse data byte received
   if ( rData == 0x00 )                          // If mouse sends it's ID byte then ...
   {
      Send2PS2Device(SET_SAMPLE_RATE);           // Send Set Sample Rate command
      WaitForACKorData();                        // Wait until mouse acknoledge received
      Send2PS2Device(SAMPLERATE_40sps);          // Send Sample to 40 samples/sec (like Windows98)
      WaitForACKorData();                        // Wait until mouse acknoledge received

      Send2PS2Device(SET_RESOLUTION);            // Send Set Resolution command
      WaitForACKorData();                        // Wait until mouse acknoledge received
      Send2PS2Device(RESOLUTION1cmm);            // Send Resolution to 1 counts/mm, best for NKC
      WaitForACKorData();                        // Wait until mouse acknoledge received

      Send2PS2Device(DATA_REPORT_ENABLE);        // Send Enable Data Reporting command
      WaitForACKorData();                        // Wait until mouse acknoledge received
   }
}

//========================================================================
//========================================================================
// Decode mouse packet (3 Byte) to mouse postion
//========================================================================

void DecodeMouse(BYTE data)
{
   BYTE mpc = mpCounter;                         // Load mouse packet counter
   BYTE mpi;

   mousePacket[mpc++] = data;                    // Store byte to mouse packet buffer
   if ( mpc > 2 )                                // Mouse packet complete ?
   {
      mpc = 0;                                   // Reset packet counter and decode packet
      mpi = mousePacket[0];
      if ( (mpi & 0x08) )                        // Must be TRUE
      {
         MouseButtons = ((mpi & 0x07) << 4);     // Mouse button states

         if ( (mpi & 0x10) )                     // Left movement
            XLpos -= mousePacket[1];
         else                                    // Right movement
            XRpos += mousePacket[1];
         if ( (mpi & 0x20) )                     // Up movement
            YUpos -= mousePacket[2];
         else                                    // Down movement
            YDpos += mousePacket[2];
      }
      else                                       // ERROR, try reset mouse
      {
         Send2PS2Device(RESET_CODE);             // Send repeat last byte command on fail
         WaitForACKorData();                     // Wait until mouse acknoledge received
      }
      ActLED_ENABLE;                       // Switch activity led on
   }
   mpCounter = mpc;                              // Restore mouse packet counter
}

//========================================================================
//========================================================================
// Output mouse position to counters
//========================================================================

void TransferData2NKC(void)
{
   BYTE temp;
   BOOL bPosDataOut = TRUE;

    // Output mouse buttons
    temp = NKCDATA_PORT;
    temp |= NKC_MOUSEBUTTONMASK;
    temp &= ~MouseButtons;
    NKCDATA_PORT = temp;
    // Output relative mouse position
    while ( bPosDataOut )
    {
       bPosDataOut = FALSE;
       // Output left position
       if ( XLpos > 0 )
       {
          XLpos--;
          NKC_MOUSELEFT_COUNT;
          bPosDataOut = TRUE;
       }
       // Output right position
       if ( XRpos > 0 )
       {
          XRpos--;
          NKC_MOUSERIGHT_COUNT;
          bPosDataOut = TRUE;
       }
       // Output up position
       if ( YUpos > 0 )
       {
          YUpos--;
          NKC_MOUSEUP_COUNT;
          bPosDataOut = TRUE;
       }
       // Output down position
       if ( YDpos > 0 )
       {
          YDpos--;
          NKC_MOUSEDOWN_COUNT;
          bPosDataOut = TRUE;
       }
       if ( bPosDataOut )
          Wait_us(12);                      // Wait 12 s, must be more than double time for latch and clear
    }
    // Signal NKC new mouse data received (-> NKC-INT\)
    NKC_INT_REQ;
}

//========================================================================
//========================================================================
// MAIN
//========================================================================

int main (void)
{
   BYTE temp;
   BOOL bRepeatSend = FALSE;                  

   STARTUP_INIT;                                 // AVR startup sequence
   PS2DATA_PORT &= ~(PS2DATA_CLK );              // Set clock to low to signal keyboard that host is buzy

   Set_Sleep_Mode(SLEEP_MODE_IDLE);              // Initialize Sleep Mode for AVR
   WDTI_IntEnable(WDT_250MS);                     // Enable Watchdog Timer Interrupt
   WDerrorSeen = FALSE;
   #if defined(TESTMODE)
   USART_Init(USART_BAUD38K);                    // Initialize USART for testing
   #endif
   ActLED_INIT;                                  // Initialize TIMER2 for activation led blinking
   ActLED_OFF;                                   // Switch LED off

   DETECT0_FALLINGEDGE;                          // Set INT0 on falling edge
   DETECT1_RISINGEDGE;                           // Set INT1 on rising edge
   ENABLE_INT0;                                  // Enable external INT0
   ENABLE_INT1;                                  // Enable external INT1

   #if defined(TESTMODE)
   USART_Transmit('*');
   #endif

   NKCDATA_PORT = NKC_MOUSEBUTTONMASK;           // No button pressed and counters to low level
   PS2DATA_PORT |= (PS2DATA_CLK | PS2DATA_OUT);  // Set clock and data to high to signal keyboard bus is idle
   PS2DATA_DDR  &= ~(PS2DATA_CLK | PS2DATA_OUT); // Set clock and data pin as input
   PS2DATA_PORT &= ~(PS2DATA_CLK | PS2DATA_OUT); // Set no pullup for clock and data pin

   SEI;                                          // Enable all used ISR

   if ( RestartCause == BV(EXTRF) )              // External reset detected at startup
   {
      Send2PS2Device(RESET_CODE);                // Send Keyboard Reset command
      WaitForACKorData();                        // Wait until keyboard acknoledge received
      WDerrorSeen = FALSE;                       // Reset WDT overflow signal
   }

   while ( TRUE )
   {
      // Wait until data byte received
      while ( PS2State != PS2STATE_DATA_RECEIVED )
      {
         SLEEP;  WDT;
         TransferData2NKC();                     // Transfer next data byte to NKC
         if ( WDerrorSeen )
         {
            WDerrorSeen = FALSE;                 // Reset WDT overflow signal
            continue;
         }
      }
      PS2State = PS2STATE_IDLE;                  // Reset State to idle.

      // Check parity bit
      temp = rParityBits & 0x01;
      if ( temp != rParity && (! bRepeatSend) )
      {
         Send2PS2Device(RESEND_CODE);            // Send repeat last packet command on fail
         bRepeatSend = TRUE;                     // Only one repeat command per packet
         mpCounter = 0;                          // Reset mouse packet counter
      }
      else
      {
         bRepeatSend = FALSE;                    // Reset repeat indicator

         if ( rData == PWRON_CODE )              // Keyboard Power On received
         {
            ActLED_OFF;                          // Switch LED off
            HandlePowerOn();
            ActLED_ON;                           // Switch LED on
            #if defined(TESTMODE)
            USART_Transmit('&');
            #endif
         }
         else
         {
            DecodeMouse(rData);                  // Decode mouse packet code to mouse position
            #if defined(TESTMODE)
            SendScanCode(rData);
            #endif
            TransferData2NKC();                  // Transfer mouse position and button data to NKC
         }
      }
   }
}

//========================================================================
